using System;
using System.Collections;
using Bespoke.Robotics.BrainStem.Exceptions;

namespace Bespoke.Robotics.BrainStem
{
    /// <summary>
    /// Represents a single BrainStem packet.
    /// </summary>
    public class Packet
    {
        /// <summary>
        /// List of available command types.
        /// </summary>
        public enum CommandCodeType
        {
            /// <summary>
            /// Heartbeat command.
            /// </summary>
            HeartBeat               = 0,

            /// <summary>
            /// 
            /// </summary>
            Route                   = 1,

			/// <summary>
			/// 
			/// </summary>
			DeviceValue				= 4,

            /// <summary>
            /// 
            /// </summary>
            GetSystemParameter      = 17,

            /// <summary>
            /// 
            /// </summary>
            SetSystemParameter      = 18,

            /// <summary>
            /// 
            /// </summary>
            SaveSystemParameters    = 19,

            /// <summary>
            /// 
            /// </summary>
            SaveServoParameters     = 20,

            /// <summary>
            /// 
            /// </summary>
            VmRun                   = 21,

            /// <summary>
            /// 
            /// </summary>
            VmKill                  = 22,

            /// <summary>
            /// 
            /// </summary>
            Debug                   = 23,

            /// <summary>
            /// 
            /// </summary>
            Reset                   = 24,

			/// <summary>
			/// 
			/// </summary>
			ReadA2D					= 25,

			/// <summary>
			/// 
			/// </summary>
			ConfigureDigitalIo		= 26,

			/// <summary>
			/// 
			/// </summary>
			DigitalIo				= 27,

			/// <summary>
			/// 
			/// </summary>
			ResetDigitalIo			= 28,

			/// <summary>
			/// 
			/// </summary>
			ReadDigitalIoTimer		= 29,

			/// <summary>
			/// 
			/// </summary>
			ReadGP2D02				= 30,

			/// <summary>
			/// 
			/// </summary>
			ConfigureServo			= 31,

			/// <summary>
			/// 
			/// </summary>
			ConfigureServoMotion	= 32,

			/// <summary>
			/// 
			/// </summary>
			PositionServoAbsolute	= 33,

			/// <summary>
			/// 
			/// </summary>
			PositionServoRelative	= 34,

			/// <summary>
			/// 
			/// </summary>
			SetServoReflex			= 35,

			/// <summary>
			/// 
			/// </summary>
			StopServo				= 36,

			/// <summary>
			/// 
			/// </summary>
			IicMasterRead			= 37,

			/// <summary>
			/// 
			/// </summary>
			SerialRelay				= 59,

            /// <summary>
            /// 
            /// </summary>
            Message                 = 128,

			/// <summary>
			/// 
			/// </summary>
			GetSystemParameterReply	= 130
        }

        /// <summary>
        /// The direction the packets are transferred.
        /// </summary>
        public enum DirectionType
        {
            /// <summary>
            /// Packets sent from host to BrainStem.
            /// </summary>
            HostToStem      = 0,

            /// <summary>
            /// Packets sent from BrainStem to host.
            /// </summary>
            StemToHost      = 1,

            /// <summary>
            /// Packets sent from a module to a router.
            /// </summary>
            ModuleToRouter  = 2
        }

        /// <summary>
        /// Maximum number of bytes that can be transmitted in the data portion of a packet.
        /// </summary>
        /// <remarks>The Packet.CommandCode takes up 1 of 8 data bytes.</remarks>
        public const int MAX_DATA_BYTES = 7;

        /// <summary>
        /// Gets the address of the IIC module referenced by the packet.
        /// </summary>
        /// <value>The address of the IIC module referenced by the packet.</value>
        public byte IicAddress
        {
            get
            {
                return mIicAddress;
            }
        }

        /// <summary>
        /// Gets the length of the data stored in the packet + 1 byte for the command.
        /// </summary>
        /// <value>The length of the data stored in the packet + 1 byte for the command</value>
        public int Length
        {
            get
            {
                return (mData == null ? 1 : mData.Length + 1);
            }
        }

        /// <summary>
        /// Gets the command being referenced by the packet.
        /// </summary>
        /// <value>The command being referenced by the packet.</value>
        public CommandCodeType CommandCode
        {
            get
            {
                return mCommandCode;
            }
        }

        /// <summary>
        /// Gets the direction of the packet.
        /// </summary>
        /// <value>The direction of the packet.</value>
        public DirectionType Direction
        {
            get
            {
                return mDirection;
            }
        }

        /// <summary>
        /// Gets the data stored in the packet.
        /// </summary>
        /// <value>The data stored in the packet.</value>
        public byte[] Data
        {
            get
            {
                return mData;
            }
        }

        /// <summary>
		/// Initializes a new instance of the Packet class.
        /// </summary>
        /// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
        /// <param name="commandCode">The command being referenced by the packet.</param>
        /// <param name="direction">The direction of the packet.</param>
        public Packet(byte iicAddress, CommandCodeType commandCode, DirectionType direction)
        {
            mIicAddress = iicAddress;
            mCommandCode = commandCode;
            mDirection = direction;
        }

        /// <summary>
		/// Initializes a new instance of the Packet class.
        /// </summary>
        /// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
        /// <param name="commandCode">The command being referenced by the packet.</param>
        /// <param name="direction">The direction of the packet.</param>
        /// <param name="data">The data stored in the packet.</param>
        public Packet(byte iicAddress, CommandCodeType commandCode, DirectionType direction, byte[] data)
            : this(iicAddress, commandCode, direction)
        {
            mData = data;
        }

        /// <summary>
        /// Send the packet to a stream.
        /// </summary>
        /// <param name="stream">The stream to send the packet on.</param>
        /// <exception cref="StreamClosedException">Thrown if the stream is not open.</exception>
        /// <exception cref="PacketDirectionException">Thrown if the specified packet's direction is not
        /// Packet.DirectionType.HostToStem.</exception>
        public virtual void Send(Stream stream)
        {
            stream.WritePacket(this);
        }

        /// <summary>
        /// Convert the packet to a byte array.
        /// </summary>
        /// <returns>A byte array containing the contents of the packet. Organized for writing
        /// to a BrainStem.</returns>
        public byte[] ToByteArray()
        {
            // 2 extra bytes for the IIC address and length
            byte[] buffer = new byte[Length + 2];
            buffer[0] = mIicAddress;
            buffer[1] = (byte)Length;
            buffer[2] = (byte)mCommandCode;
            if (mData != null)
            {
                Array.Copy(mData, 0, buffer, 3, mData.Length);
            }

            return buffer;
        }

        /// <summary>
        /// The address of the IIC module referenced by the packet.
        /// </summary>
        protected byte mIicAddress;
        
        /// <summary>
        /// The command being referenced by the packet.
        /// </summary>
        protected CommandCodeType mCommandCode;

        /// <summary>
        /// The direction of the packet.
        /// </summary>
        protected DirectionType mDirection;

        /// <summary>
        /// The data stored in the packet.
        /// </summary>
        protected byte[] mData;
    }

	/// <summary>
	/// Represents a collection of BrainStem packets.
	/// </summary>
	public class PacketCollection : ICollection
	{
		/// <summary>
		/// Gets or sets the element at the specified index.
		/// </summary>
		/// <param name="index">The zero-based index of the element to get or set.</param>
		public Packet this[int index]
		{
			get
			{
				return (Packet)mArrayList[index];
			}
			set
			{
				mArrayList[index] = value;
			}
		}

		/// <summary>
		/// Gets the number of elements contained in the collection.
		/// </summary>
		public int Count
		{
			get
			{
				return mArrayList.Count;
			}
		}

		/// <summary>
		/// Gets a value indicating whether access to the collection is synchronized (thread safe).
		/// </summary>
		public bool IsSynchronized
		{
			get
			{
				return mArrayList.IsSynchronized;
			}
		}

		/// <summary>
		/// Gets an object that can be used to synchronize access to the collection.
		/// </summary>
		public object SyncRoot
		{
			get
			{
				return mArrayList.SyncRoot;
			}
		}

		/// <summary>
		/// Initializes a new instance of the PacketCollection class.
		/// </summary>
		public PacketCollection()
		{
			mArrayList = new ArrayList();
		}

		/// <summary>
		/// Initializes a new instance of the PacketCollection class.
		/// </summary>
		/// <param name="packetArray">An array of packets to populate the collection with.</param>
		public PacketCollection(Packet[] packetArray)
		{
			foreach (Packet packet in packetArray)
			{
				Add(packet);
			}
		}

		/// <summary>
		/// Adds a packet to the end of the collection.
		/// </summary>
		/// <param name="packet">The <see cref="Packet"/> to be added to the end of the collection.</param>
		/// <returns>The index at which the value has been added.</returns>
		public int Add(Packet packet)
		{
			return mArrayList.Add(packet);
		}

		/// <summary>
		/// Removes all elements from the collection.
		/// </summary>
		public void Clear()
		{
			mArrayList.Clear();
		}

		/// <summary>
		/// Determines whether an element is in the collection.
		/// </summary>
		/// <param name="packet">The <see cref="Packet"/> to locate in the collection.</param>
		/// <returns><b>true</b> if item is found in the collection; otherwise, <b>false</b>.</returns>
		public bool Contains(Packet packet)
		{
			return mArrayList.Contains(packet);
		}

		/// <summary>
		/// Copies the entire collection to a compatible one-dimensional Array,
		/// starting at the beginning of the target array.
		/// </summary>
		/// <param name="array">The one-dimensional Array that is the destination of
		/// the elements copied from the collection. The Array must have zero-based indexing.</param>
		/// <param name="arrayIndex">The zero-based index in <paramref name="array"/> at which copying begins.</param>
		public void CopyTo(Array array, int arrayIndex)
		{
			mArrayList.CopyTo(array, arrayIndex);
		}

		/// <summary>
		/// Returns an enumerator that iterates through the collection.
		/// </summary>
		/// <returns>An <see cref="IEnumerator"/> for the entire collection.</returns>
		public IEnumerator GetEnumerator()
		{
			return mArrayList.GetEnumerator();
		}

		/// <summary>
		/// Returns the zero-based index of the first occurrence of a value in the
		/// collection or in a portion of it.
		/// </summary>
		/// <param name="packet">The <see cref="Packet"/> to locate in the collection.</param>
		/// <returns>The zero-based index of the first occurrence of <paramref name="packet"/> within the entire collection,
		/// if found; otherwise, -1. </returns>
		public int IndexOf(Packet packet)
		{
			return mArrayList.IndexOf(packet);
		}

		/// <summary>
		/// Returns the zero-based index of the first occurrence of a value in the
		/// collection or in a portion of it. 
		/// </summary>
		/// <param name="commandCode">The <see cref="Packet.CommandCodeType"/> to locate in the collection.</param>
		/// <returns>The zero-based index of the first occurrence of <paramref name="commandCode"/> within the entire collection,
		/// if found; otherwise, -1. </returns>
		public int IndexOf(Packet.CommandCodeType commandCode)
		{
			int index = -1;

			for (int i = 0; i < mArrayList.Count; i++)
			{
				if (this[i].CommandCode == commandCode)
				{
					index = i;
					break;
				}
			}

			return index;
		}

		/// <summary>
		/// Returns the zero-based index of the last occurrence of a value in the collection or in a portion of it.
		/// </summary>
		/// <param name="packet">The <see cref="Packet"/> to locate in the collection.</param>
		/// <returns>The zero-based index of the last occurrence of <paramref name="packet"/> within the entire the collection,
		/// if found; otherwise, -1.</returns>
		public int LastIndexOf(Packet packet)
		{
			return mArrayList.LastIndexOf(packet);
		}

		/// <summary>
		/// Returns the zero-based index of the last occurrence of a value in the collection or in a portion of it.
		/// </summary>
		/// <param name="commandCode">The <see cref="Packet.CommandCodeType"/> to locate in the collection.</param>
		/// <returns>The zero-based index of the last occurrence of <paramref name="commandCode"/> within the entire the collection,
		/// if found; otherwise, -1.</returns>
		public int LastIndexOf(Packet.CommandCodeType commandCode)
		{
			int index = -1;

			for (int i = 0; i < mArrayList.Count; i++)
			{
				if (this[i].CommandCode == commandCode)
				{
					index = i;
				}
			}

			return index;
		}

		/// <summary>
		/// Removes the first occurrence of a specific object from the collection. 
		/// </summary>
		/// <param name="packet">The <see cref="Packet"/> to remove from the collection.</param>
		public void Remove(Packet packet)
		{
			mArrayList.Remove(packet);
		}

		/// <summary>
		/// Removes the element at the specified index of the collection.
		/// </summary>
		/// <param name="index">The zero-based index of the element to remove.</param>
		public void RemoveAt(int index)
		{
			mArrayList.RemoveAt(index);
		}

		private ArrayList mArrayList;
	}
}
